from __future__ import print_function
#
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
#
#   Copyright (C) 2011 Mirna Lerotic, 2nd Look
#   http://2ndlookconsulting.com
#   License: GNU GPL v3
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.

import numpy as np
import os


title = 'text spectrum'
extension = ['*.csv','*.txt','*.xas']
read_types = ['spectrum']
write_types = ['spectrum','stack']


def identify(filename):
    try:
        with open(file_name, 'tr') as check_file:  # try open file in text mode
            check_file.read()
            return True
    except:
        return False

def GetFileStructure(FileName):
    return None


#----------------------------------------------------------------------
def read(filename, self, selection=None, *args, **kwargs):

    f = open(str(filename),'r')
    
    elist = []
    ilist = []    

    for line in f:
        if line.startswith(("*","%","#")):
            pass
        else:
            e, i = [float (x) for x in line.split(',')] 
            elist.append(e)
            ilist.append(i)
            
    self.evi0 = np.array(elist)
    self.i0data = np.array(ilist) 
            
    f.close()
    
    self.i0_dwell = None
       
    return
    
#----------------------------------------------------------------------
def write(filename, data_object, data_type):
    """Switchyard for writing different types of data."""
    if data_type in ['spectrum']:
        write_spectrum(filename, data_object.absdata, data_object.ev)
    elif data_type in ['stack']:
        write_spectrum(filename, np.average(data_object.absdata,axis=(0,1)), energies=data_object.ev, title='I0 Spectrum')

#----------------------------------------------------------------------
def write_spectrum(filename, data, energies, title=None ):
    if title is None:
        title = 'Spectrum'
    
    with open(filename, 'w') as f:
        print('********************* '+title+'  ********************', file=f)
        print('*', file=f)
        print('* ev, intensity', file=f)
        for ie in range(len(energies)):
            print('{0:06.2f}, {1:06f}'.format(energies[ie], data[ie]), file=f)






# -*- coding: utf-8 -*-
#
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
#
#   Copyright (C) 2015 Benjamin Watts, Paul Scherrer Institut
#   License: GNU GPL v3
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.


from __future__ import print_function
import json
import re, numpy, sys
from os.path import splitext, join, dirname, isfile
from collections import OrderedDict


title = 'SDF'
extension = ['*.hdr']
read_types = ['spectrum','image','stack']
write_types = []
def identify(filename):
    try:
        if isfile(splitext(filename)[0] + '.json'):
            print("JSON-file found. No need to fetch again.")
            return False
        else:
            HDR = HDR_FileParser(filename)
            return HDR.num_regions > 0  # return true if file contains at least one region
    except:
        return False

def GetFileStructure(FileName):
    HDR = HDR_FileParser(FileName)
    if HDR.num_regions<2 and HDR.num_channels<2 and isfile(splitext(FileName)[0] + '.json'): #if json file exists, skip datachoicedialog
        return None # exit if only one choice
    D = OrderedDict()
    for i,R in enumerate(['Region_'+str(r) for r in range(HDR.num_regions)]):
        D[R] = OrderedDict()
        D[R].definition = 'SDF'
        D[R].scan_type = HDR.hdr['ScanDefinition']['Type']
        D[R].data_shape = HDR.data_size[i]
        D[R].data_axes = [HDR.hdr['ScanDefinition']['Regions'][1]['PAxis']['Name'],HDR.hdr['ScanDefinition']['Regions'][1]['QAxis']['Name'],HDR.hdr['ScanDefinition']['StackAxis']['Name']]
        for ch in range(1,HDR.num_channels+1):
            D[R][HDR.hdr['Channels'][ch]['Name']] = OrderedDict()
    return D



#----------------------------------------------------------------------
class HDR_FileParser:
  """Parse .hdr file for metadata."""
  hdr = []
  f = []
  def __init__(self, fileName,identify=False):
    if not HDR_FileParser.hdr or HDR_FileParser.f != fileName: # prevent class from fetching and parsing the *.hdr file multiple times. Use the class attribute "hdr" instead if available and check if a new file (not "f") is loaded.
        # compile some regular expressions
        self.MatchReStruct = re.compile('[\s\{\}\(\)=";]')
        self.MatchReArray = re.compile('[,\s\{\(\);]')
        self.__file = open(fileName)
        # Parse the HDR file
        HDR_FileParser.hdr = self.parseStructure()
        HDR_FileParser.f = fileName
        self.__file.close()
    else:
        None
    self.hdr = HDR_FileParser.hdr
    if 'ScanDefinition' in self.hdr:
      self.num_regions = int(self.hdr['ScanDefinition']['Regions'][0])
      self.num_channels = int(self.hdr['Channels'][0])
      self.file_path, self.file_ext = splitext(fileName)
      self.data_size = self.parse_DataSize()
      self.data_names = self.parseDataNames()
    else:
      self.num_regions = 0
      self.num_channels = 0



#----------------------------------------------------------------------
  def parseStructure(self):
    """.hdr files consist of structures and arrays. This routine sorts through the structure parts."""
    Structure = {}
    BuildWord=''
    BeforeEq=True
    QuotedWord=False
    raw = self.__file.read(1)
    while len(raw) > 0:#until we reach the end of the file
      matched = self.MatchReStruct.match(raw)
      if matched == None:
        BuildWord+=raw
      elif matched.group() == '"':
        QuotedWord= not QuotedWord
      elif QuotedWord==True:
        BuildWord+=raw
      elif matched.group() == '=':
        FieldName=BuildWord
        BuildWord=''
        BeforeEq=False
      elif matched.group() == ';':
        try:  # convert numbers into ints or floats or end up with a string
            Structure[FieldName] = int(BuildWord)
        except ValueError:
            try:
                Structure[FieldName] = float(BuildWord)
            except ValueError:
                Structure[FieldName] = BuildWord
        except TypeError:
            Structure[FieldName] = BuildWord
        BuildWord=''
        BeforeEq=True
      elif matched.group() == '{':
        #Must be after =
        BuildWord = self.parseStructure()
      elif matched.group() == '}':
        #break loop and return dictionary
        break
      elif matched.group() == '(':
        #Must be after =
        BuildWord= self.parseArray()
      elif matched.group() == ')':
        #This should not happen
        print(') in structure')
      raw = self.__file.read(1)
    return Structure

#----------------------------------------------------------------------
  def parseArray(self):
    """.hdr files consist of structures and arrays. This rountine sorts through the array parts."""
    Array = []
    BuildWord=''
    raw = self.__file.read(1)
    while len(raw) > 0:#until we reach the end of the file
      matched = self.MatchReArray.match(raw)
      if matched == None:
        BuildWord+=raw
      elif matched.group() == ',':
        if len(BuildWord) > 0:
            try:    #convert numbers in array into ints or floats
                Array.append(int(BuildWord))
            except ValueError:
                Array.append(float(BuildWord))
            BuildWord=''
      elif matched.group() == ';':
        print('; in array')
      elif matched.group() == '{':
        Array.append(self.parseStructure())
      elif matched.group() == '(':
        Array.append(self.parseArray())
      elif matched.group() == ')':
        if len(BuildWord) > 0:
            try:  #convert numbers in array into ints or floats
                Array.append(int(BuildWord))
            except ValueError:
                try:
                    Array.append(float(BuildWord))
                except ValueError: # There is an error in the HDF5toSDF conversion script at the SLS/PolLux which appends arrays with a superfluous "}". Obviously, this char cannot be converted to int or float. We therefore just skip it here.
                    pass
        break
      raw = self.__file.read(1)
    return Array

#----------------------------------------------------------------------
  def parseDataNames(self): #ToDo: Assigning Scan types by DataFlags is prone to errors. Better detect automatically via (HDR.hdr['ScanDefinition']['Regions'][0]) etc.
    """Figure out names for the .xsp or .xim files that contain the actual data, then check that the files actually exist, printing warnings if they don't."""
    DataNames = []#Regions
    DataNames2 = []#Channels
    DataNames3 = []#Energies
    Alphabet = 'abcdefghijklmnopqrstuvwxyz'
    DataFlag = self.hdr['ScanDefinition']['Flags']
    if DataFlag in ['Spectra','Multi-Region Spectra']:
      for num_R in range(self.num_regions):
        DataNames2 = []
        for num_Ch in range(self.num_channels):
          DataNames2.append([self.file_path+'_'+str(num_R)+'.xsp'])
        DataNames.append(DataNames2)
    elif DataFlag == 'Image':
      DataNames2 = []
      for num_Ch in range(self.num_channels):
        DataNames2.append([self.file_path+'_'+Alphabet[num_Ch]+'.xim'])
      DataNames.append(DataNames2)
    elif DataFlag in ['Multi-Region Image']:
      for num_R in range(self.num_regions):
        DataNames2 = []
        for num_Ch in range(self.num_channels):
          DataNames2.append([self.file_path+'_'+Alphabet[num_Ch]+str(num_R)+'.xim'])
        DataNames.append(DataNames2)
    elif DataFlag in ['Multi-Region Image Stack']:
      for num_Ch in range(self.num_channels):
        DataNames2 = []
        for num_R in range(self.num_regions):
            DataNames3 = []
            for num_E in range(self.data_size[0][2]):
                DataNames3.append(self.file_path + '_' + Alphabet[num_Ch] + str(num_E).zfill(3)+str(num_R)+ '.xim')
            DataNames2.append(DataNames3)
        DataNames = [DataNames2]
    elif DataFlag == 'Image Stack':
      DataNames2 = []
      for num_Ch in range(self.num_channels):
        DataNames3 = []
        for num_E in range(self.data_size[0][2]):
          DataNames3.append(self.file_path+'_'+Alphabet[num_Ch]+str(num_E).zfill(3)+'.xim')
        DataNames2.append(DataNames3)
      DataNames = [DataNames2]
    else:
      print("WARNING! Unknown flag:", DataFlag)
    #for num_R in range(len(DataNames)):#Check that names correspond to existing files
      #for num_Ch in range(len(DataNames[num_R])):
        #for num_E in range(len(DataNames[num_R][num_Ch])):
          #if exists(DataNames[num_R][num_Ch][num_E]) == False:
            #print "WARNING! Data file doesn't exist:", DataNames[num_R][num_Ch][num_E]
    return DataNames

#----------------------------------------------------------------------
  def parse_DataSize(self):
    """Calculate data array size. This is useful for making sure all of the lists of data are the correct length."""
    DataSize = []
    for R_num in range(self.num_regions):
      DataSize.append([1,1,1])# [PAxis,QAxis,StackAxis] (switch to [X1,X2,E] later)]
      DataSize[R_num][0] = int(self.hdr['ScanDefinition']['Regions'][R_num+1]['PAxis']['Points'][0])
      if 'QAxis' in self.hdr['ScanDefinition']['Regions'][R_num+1] and 'Points' in self.hdr['ScanDefinition']['Regions'][R_num+1]['QAxis']:
        DataSize[R_num][1] = int(self.hdr['ScanDefinition']['Regions'][R_num+1]['QAxis']['Points'][0])
      if 'StackAxis' in self.hdr['ScanDefinition'] and 'Points' in self.hdr['ScanDefinition']['StackAxis']:
        DataSize[R_num][2] = int(self.hdr['ScanDefinition']['StackAxis']['Points'][0])
      if self.hdr['ScanDefinition']['Type'] in ['NEXAFS Point Scan','NEXAFS Line Scan']:
        DataSize[R_num] = [DataSize[R_num][1],1,DataSize[R_num][0]]#switch to [X1,X2,E] format
#        DataSize[R_num] = [1,DataSize[R_num][1],DataSize[R_num][0]]#also works, but might be problematic for finding number of spatial points
    return DataSize




#----------------------------------------------------------------------
#----------------------------------------------------------------------
def read(filename, self, selection=None, JSONstatus=None):
    HDR = HDR_FileParser(filename)
    if JSONstatus:
        with open(splitext(filename)[0] + '.json', 'w') as outfile:
            json.dump(HDR.hdr, outfile, indent=4, sort_keys=True, ensure_ascii=True)
            print("JSON-file written at "+ splitext(filename)[0] + '.json')
    allowed =['Image Stack','Image','Multi-Region Image Stack']
    if HDR.hdr['ScanDefinition']['Flags'] in allowed:

        self.x_dist = numpy.array([float(i) for i in HDR.hdr['ScanDefinition']['Regions'][selection[0]+1]['PAxis']['Points'][1:] ])
        self.y_dist = numpy.array([float(i) for i in HDR.hdr['ScanDefinition']['Regions'][selection[0]+1]['QAxis']['Points'][1:] ])
        self.ev = numpy.array([float(i) for i in HDR.hdr['ScanDefinition']['StackAxis']['Points'][1:] ])

        self.n_cols = len(self.x_dist)
        self.n_rows = len(self.y_dist)
        self.n_ev = len(self.ev)

        msec = float(HDR.hdr['ScanDefinition']['Dwell'])
        self.data_dwell = numpy.ones((self.n_ev))*msec

        imagestack = numpy.empty((self.n_cols,self.n_rows,self.n_ev), numpy.int32)
        for i in range(len(HDR.data_names[selection[1]][selection[0]])):
            try:
                imagestack[:,:,i] = numpy.loadtxt(HDR.data_names[selection[1]][selection[0]][i], numpy.int32).T
            except ValueError:
                print("Aborted stack or XIMs with inconsistent dimensions.")
                imagestack[:,:,i] = numpy.nan
            except IOError:
                print("Image file not found.")
                imagestack[:,:,i] = numpy.nan
        self.absdata = numpy.empty((self.n_cols,self.n_rows, self.n_ev))

        self.absdata = numpy.reshape(imagestack, (self.n_cols,self.n_rows, self.n_ev), order='F')

        self.fill_h5_struct_from_stk()

    else:
        print("Unknown Format")

    return

#----------------------------------------------------------------------
def read_sdf_i0(self, filename):
    HDR = HDR_FileParser(filename)

    if 'ScanType' in HDR.hdr['ScanDefinition'] and HDR.hdr['ScanDefinition']['ScanType'] == 'Spectra':
        Energies = HDR.hdr['ScanDefinition']['Regions'][1]['PAxis']['Points'][1:]
        tempimage = numpy.loadtxt(HDR.data_names[0][0][0], numpy.float32)
        Data = tempimage[:,1]
    elif HDR.hdr['ScanDefinition']['Type'] == 'NEXAFS Line Scan':
        Energies = HDR.hdr['ScanDefinition']['Regions'][1]['PAxis']['Points'][1:]
        tempimage = numpy.loadtxt(HDR.data_names[0][0][0], numpy.int32)
        Data = numpy.mean(tempimage,axis=0)
    else:# Image Stack
        Energies = HDR.hdr['ScanDefinition']['StackAxis']['Points'][1:]
        tempimage = numpy.empty((HDR.data_size[0][0],HDR.data_size[0][1]), numpy.int32)
        Data = numpy.empty((HDR.data_size[0][2]), numpy.int32)
        for i in range(len(HDR.data_names[0][0])):
            tempimage = numpy.loadtxt(HDR.data_names[0][0][i], numpy.int32)
            Data[i] = numpy.mean(tempimage)

    msec = float(HDR.hdr['ScanDefinition']['Dwell'])#shouldn't this be needed?
    self.i0_dwell = msec
    self.evi0 = numpy.array([float(i) for i in Energies])
    self.i0data = Data
    return


# -*- coding: utf-8 -*-
#
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
#
#   Copyright (C) 2015 Benjamin Watts, Paul Scherrer Institute
#   License: GNU GPL v3
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.

"""
The file_plugins system is exposed to general code through the functions defined here in __init__.py:

identify(filename)                  : Returns an instance of the plugin that claims to deal with the file at the URL 'filename'.
GetFileStructure(filename)          : Returns a structure describing the internal organisation of the file, indicating sets of data available to choose from.
load(filename,stack_object,..)      : Loads data from the URL 'filename' into the object (data_stack type) 'stack_object'. The plugin used can be stated or determined automatically, using 'identify'.

Further functions for writing files via the plugins need to be written yet. To access the system, you should import the module ('import file_plugins') and then access the above functions as attributes of the module (e.g. 'file_plugins.load('data.hdf5',data_stk)' ).

Each file plugin should be included here in the 'file_plugins' directory. Each plugin should define the following:

title                           : A short string naming the plugin.
extension                       : A list of strings indicating the file extensions that the plugin handles (e.g. ['*.txt']).
read_types                      : A list of strings indicating the data types that the plugin will read (e.g. ['spectrum','image','stack']).
write_types                     : A list of strings indicating the data types that the plugin will write (e.g. ['spectrum','image','stack']).
identify(filename)              : Returns boolean indicating if the plugin can read the file at URL 'filename'.
GetFileStructure(filename)      : Returns a structure describing the internal organisation of the file, indicating sets of data available to choose from.
read(filename,stack_object,..)  : Loads data from the URL 'filename' into the object (data_stack type) 'stack_object'.

"""
from __future__ import print_function

import pkgutil, imp, os, sys
import numpy
from .. import data_stack

verbose = True

# These variables declare the options that each plugin can claim the ability to handle
actions = ['read','write']
data_types = ['spectrum','image','stack','results']

# Go through the directory and try to load each plugin
plugins = []

for m in pkgutil.iter_modules(path=__path__):
    if verbose: print("Loading file plugin:", m[1], ".", end=' ')
    try:
        details = imp.find_module(m[1],__path__)
        # check if there is a read() function in plugin
        if 'read' in dir(imp.load_module(m[1],*details)):
            plugins.append(imp.load_module(m[1],*details))
            if verbose: print("("+plugins[-1].title+") Success!")
        else:
            if verbose: print('Not a valid plugin - skipping.')

    except ImportError as e:
        if verbose: print("prerequisites not satisfied:", e)


# if getattr(sys, 'frozen', False):
#     module_names = ['file_dataexch_hdf5', 'file_ncb', 'file_nexus_hdf5', 'file_sdf', 'file_stk', 'file_tif', 'file_xrm']
#     for m in module_names:
#         if verbose: print "Loading file plugin:", m, "...",
#         try:
#
#
#             details = imp.find_module(m)
#             # check if there is a read() function in plugin
#             if 'read' in dir(imp.load_module(m,*details)):
#                 plugins.append(imp.load_module(m,*details))
#                 if verbose: print "("+plugins[-1].title+") Success!"
#             else:
#                 if verbose: print 'Not a valid plugin - skipping.'
#
#         except ImportError as e:
#             if verbose: print "prerequisites not satisfied:", e




# Go through set of plugins and assemble lists of supported file types for each action and data type
supported_filters = dict([a,dict([t,[]] for t in data_types)] for a in actions)
supported_plugins = dict([a,dict([t,[]] for t in data_types)] for a in actions)
filter_list = dict([a,dict([t,[]] for t in data_types)] for a in actions)
for P in plugins:
    for action in actions:
        for data_type in data_types:
            if data_type in getattr(P,action+'_types'):
                filter_list[action][data_type].append(P.title+' ('+' '.join(P.extension)+')')
                supported_plugins[action][data_type].append(P)
                for ext in P.extension:
                    if ext not in supported_filters[action][data_type]:
                        supported_filters[action][data_type].append(ext)
for data_type in data_types:
    filter_list['read'][data_type] = ['Supported Formats ('+' '.join(supported_filters['read'][data_type])+')']+filter_list['read'][data_type]
    filter_list['read'][data_type].append('All files (*.*)')


def load(filename, stack_object=None, plugin=None, selection=None, json=None):
    """
    Pass the load command over to the appropriate plugin so that it can import data from the named file.
    """
    if plugin is None:
        plugin = identify(filename)
    if plugin is None:
        return None
    else:
        print("load", filename, "with the", plugin.title, "plugin.")
        if selection is None:
            plugin.read(filename, stack_object, selection, json)
        elif len(selection) == 1:
            plugin.read(filename, stack_object, selection[0], json)
        else:
            plugin.read(filename,stack_object,selection[0],json)
            temp_stack = data_stack.data(stack_object.data_struct)
            full_stack = stack_object.absdata.copy()
            for s in selection[1:]:
                plugin.read(filename,temp_stack,s)
                if full_stack.shape[1] > temp_stack.absdata.shape[1]:
                    temp_stack.absdata = numpy.pad(temp_stack.absdata,((0,0),(0,full_stack.shape[1]-temp_stack.absdata.shape[1]),(0,0)), mode='constant',constant_values=0)
                elif full_stack.shape[1] < temp_stack.absdata.shape[1]:
                    full_stack = numpy.pad(full_stack,((0,0),(0,temp_stack.absdata.shape[1]-full_stack.shape[1]),(0,0)), mode='constant',constant_values=0)
                full_stack = numpy.vstack((full_stack,temp_stack.absdata))
            stack_object.absdata = full_stack
            stack_object.x_dist = numpy.arange(full_stack.shape[0])
            stack_object.y_dist = numpy.arange(full_stack.shape[1])
            stack_object.n_cols = len(stack_object.x_dist)
            stack_object.n_rows = len(stack_object.y_dist)
            return

def save(filename, data_object, data_type, plugin=None):
    """
    Pass the save command over to the appropriate plugin so that it can write data to the named file.
    """
    print("save", filename, "with the", plugin.title, "plugin.")
    plugin.write(filename, data_object, data_type)

def GetFileStructure(filename, plugin=None):
    """
    Use the plugin to skim-read the file and return the structure of the data.
    Returns None if there is only a single data array (i.e. no choices to be made).
    """
    if plugin is None:
        plugin = identify(filename)
    if plugin is None:
        return None
    else:
        print("get info from", filename, "with the", plugin.title, "plugin.")
        FileInfo = plugin.GetFileStructure(filename)
        #if FileInfo is not None:
            #print len(FileInfo), len(FileInfo[next(iter(FileInfo))])
            #print FileInfo
        return FileInfo

def identify(filename):
    """
    Cycle through plugins until finding one that claims to understand the file format.
    First it tries those claiming corresponding file extensions, followed by all other plugins until an appropriate plugin is found.
    """
    print("Identifying file:", filename, "...", end=' ')
    ext = os.path.splitext(filename)[1]
    flag = [True]*len(plugins)
    for i,P in enumerate(plugins):
        if '*'+ext in P.extension:
            if P.identify(filename):
                return P
            elif flag[i] == True: #if plugin returns False, e.g. dataexch_hdf5 does not match, try the next plugin and find the same extension
                flag[i] = False
                continue
            else:
                break
    print("Error! unknown file type.")
    return None




# 
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
# 
#   Copyright (C) 2015 Mirna Lerotic, 2nd Look
#   http://2ndlookconsulting.com
#   License: GNU GPL v3
#
#   Originally part of tomotv, code licensed under BSD license.
#   Website: https://github.com/emmanuelle/tomo-tv
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import numpy as np
from scipy import sparse
from scipy import ndimage
from scipy import fftpack
from ._rank_order import rank_order

# --------------- Tomo projection operator  --------------------

def build_projection_operator(l_x, n_dir=None, angles=None, l_det=None, subpix=1,
                                offset=0, pixels_mask=None):
    """
    Compute the tomography design matrix.

    Parameters
    ----------

    l_x : int
        linear size of image array

    n_dir : int, default l_x
        number of angles at which projections are acquired. n_dir
        projection angles are regularly spaced between 0 and 180.

    l_det : int, default is l_x
        number of pixels in the detector. If l_det is not specified,
        we suppose that l_det = l_x.

    subpix : int, default 1
        number of linear subdivisions used to compute the projection of
        one image pixel onto a detector pixel. For example, if subpix=2,
        one image pixel is divided into 2x2 subpixels that are projected
        onto the detector, and the value of the projections is computed
        from these 4 projections.

    offset : int, default 0
        width of the strip of image pixels not covered by the detector. 
        offset > 0 means that the image is acquired in local tomography
        (aka ROI) mode, with the image larger than the detector. If the
        linear size of the array is l_x, the size of the detector is 
        l_x - 2 offset.

    pixels_mask : 1-d ndarray of size l_x**2
        mask of pixels to keep in the matrix (useful if one wishes
        to remove pixels inside or outside of a circle, for example)

    Returns
    -------
    p : sparse matrix of shape (n_dir l_x, l_x**2), in csr format
        Tomography design matrix. The csr (compressed sparse row)
        allows for efficient subsequent matrix multiplication. The
        dtype of the elements is float32, in order to save memory.

    Notes
    -----
    The returned matrix is sparse, but may nevertheless require a lot
    of memory for large l_x. For example, with l_x=512 and n_dir=512,
    the operator takes around 3 Gb of memory. The memory cost is of
    the order of l_x^2 x n_dir x 8 in bytes.

    For a given angle, the center of the pixels are rotated by the
    angle, and projected onto the detector. The value of the data pixel
    is added to the two pixels of the detector in between which the
    projection is located, with weights determined by a linear
    interpolation.

    Using subpix > 1 slows down the computation of the operator, because
    a histogram in 2-D has to be computed in order to group the projections 
    of subpixels corresponding to a single image pixel.
    (this should be accelerated by a Cython function... to be written)

    Examples
    --------
    >>> # Image with 256 pixels, 128 directions 
    >>> op = build_projection_operator(256, n_dir=128)

    >>> # Image with 128 pixels (to be reconstructed), 256 detector pixels
    >>> # subpix = 2 is used for a good precision of the projection of the
    >>> # coarse image pixels
    >>> op = build_projection_operator(128, n_dir=256, l_det=256, subpix=2)
    
    >>> # Image with 256 pixels, that is twice the size of the detector that
    >>> # has 128 pixels.
    >>> op = build_projection_operator(256, n_dir=256, l_det=128, offset=64)
    
    >>> # Image with 256 pixels, that is twice the size of the detector that
    >>> # has 256 pixels. We use subpixels for better precision.
    >>> op = build_projection_operator(256, n_dir=256, l_det=256, offset=64)
    
    >>> # Using a mask: projection operator only for pixels inside a
    >>> # central circle
    >>> l_x = 128
    >>> X, Y = np.ogrid[:l_x, :l_x]
    >>> mask = (X - l_x/2)**2 + (Y - l_x/2)**2 < (l_x/2)**2
    >>> op = build_projection_operator(l_x, pixels_mask=mask)
    >>> op.shape
    (16384, 12849)
    """
    if l_det is None:
        l_det = l_x
    X, Y = _generate_center_coordinates(subpix*l_x)
    X *= 1./subpix
    Y *= 1./subpix
    Xbig, Ybig = _generate_center_coordinates(l_det)
    Xbig *= (l_x - 2*offset) / float(l_det)
    orig = Xbig.min()
    labels = None
    if subpix > 1:
        # Block-group subpixels
        Xlab, Ylab = np.mgrid[:subpix * l_x, :subpix * l_x]
        labels = (l_x * (Xlab / subpix) + Ylab / subpix).ravel()
    if n_dir is None:
        n_dir = l_x
    if angles is None:
        angles = np.linspace(0, np.pi, n_dir, endpoint=False)
    else:
        n_dir = len(angles)
    weights, data_inds, detector_inds = [], [], []
    # Indices for data pixels. For each data, one data pixel
    # will contribute to the value of two detector pixels.
    for i, angle in enumerate(angles):
        # rotate data pixels centers
        Xrot = np.cos(angle) * X - np.sin(angle) * Y
        # compute linear interpolation weights
        inds, dat_inds, w = _weights_fast(Xrot, dx=(l_x - 2*offset)/float(l_det),
                                    orig=orig, labels=labels)
        # crop projections outside the detector
        mask = np.logical_and(inds >= 0, inds < l_det)
        weights.append(w[mask])
        detector_inds.append((inds[mask] + i * l_det).astype(np.int32))
        data_inds.append(dat_inds[mask])
    weights = np.concatenate(weights)
    weights /= subpix**2
    detector_inds = np.concatenate(detector_inds)
    data_inds = np.concatenate(data_inds)
    if pixels_mask is not None:
        if pixels_mask.ndim > 1:
            pixels_mask = pixels_mask.ravel()
        mask = pixels_mask[data_inds]
        data_inds = data_inds[mask]
        data_inds = rank_order(data_inds)[0]
        detector_inds = detector_inds[mask]
        weights = weights[mask]
    proj_operator = sparse.coo_matrix((weights, (detector_inds, data_inds)))
    return sparse.csr_matrix(proj_operator)

def _weights_fast(x, dx=1, orig=0, ravel=True, labels=None):
    """
    Compute linear interpolation weights for projection array `x`
    and regularly spaced detector pixels separated by `dx` and
    starting at `orig`.
    """
    if ravel:
        x = np.ravel(x)
    floor_x = np.floor((x - orig) / dx).astype(np.int32)
    alpha = ((x - orig - floor_x * dx) / dx).astype(float32)
    inds = np.hstack((floor_x, floor_x + 1))
    weights = np.hstack((1 - alpha, alpha))
    data_inds = np.arange(x.size, dtype=np.int32)
    data_inds = np.hstack((data_inds, data_inds))
    if labels is not None:
        data_inds = np.hstack((labels, labels))
        order = np.argsort(inds)
        inds, data_inds, weights = inds[order], data_inds[order], weights[order]
        steps = np.nonzero(np.diff(inds) > 0)[0] + 1
        steps = np.concatenate(([0], steps))
        inds_s, data_inds_s, weights_s = [], [], []
        for i in range(len(steps) - 1):
            d, w = data_inds[steps[i]:steps[i+1]], \
                      weights[steps[i]:steps[i+1]]
            count = np.bincount(d, weights=w)
            mask = count>0
            w = count[mask]
            weights_s.append(w)
            datind = np.arange(len(mask))[mask] 
            data_inds_s.append(datind)
            detind = inds[steps[i]]*np.ones(mask.sum()) 
            inds_s.append(detind)
            #stop
        inds = np.concatenate(inds_s)
        data_inds = np.concatenate(data_inds_s)
        weights = np.concatenate(weights_s)
    return inds, data_inds, weights

def _weights(x, dx=1, orig=0, ravel=True, labels=None):
    """
    Compute linear interpolation weights for projection array `x`
    and regularly spaced detector pixels separated by `dx` and
    starting at `orig`.
    """
    if ravel:
        x = np.ravel(x)
    floor_x = np.floor((x - orig) / dx).astype(np.int32)
    alpha = ((x - orig - floor_x * dx) / dx).astype(float32)
    inds = np.hstack((floor_x, floor_x + 1))
    weights = np.hstack((1 - alpha, alpha))
    data_inds = np.arange(x.size, dtype=np.int32)
    data_inds = np.hstack((data_inds, data_inds))
    if labels is not None:
        data_inds = np.hstack((labels, labels))
        w = np.histogram2d(data_inds, inds,
             bins=(np.arange(data_inds.max()+1.5), np.arange(inds.max()+1.5)),
             weights=weights)[0]
        data_inds, inds = np.argwhere(w>0).T
        weights = w[w>0]
    return inds, data_inds, weights

def _weights_nn(x, dx=1, orig=0, ravel=True):
    """
    Nearest-neighbour interpolation
    """
    if ravel:
        x = np.ravel(x)
    floor_x = np.floor(x - orig)
    return floor_x.astype(float32)



def _generate_center_coordinates(l_x):
    """
    Compute the coordinates of pixels centers for an image of
    linear size l_x
    """
    l_x = float(l_x)
    X, Y = np.mgrid[:l_x, :l_x]
    center = l_x / 2.
    X += 0.5 - center
    Y += 0.5 - center
    return X, Y



# ----------------- Direct projection method -------------------------
# (without computing explicitely the design matrix)


def back_projection(projections):
    """
    Back-projection (without filtering)

    Parameters
    ----------
    projections: ndarray of floats, of shape n_dir x l_x
        Each line of projections is the projection of a data image
        acquired at a different angle. The projections angles are
        supposed to be regularly spaced between 0 and 180.

    Returns
    -------
    recons: ndarray of shape l_x x l_x
        Reconstructed array

    Notes
    -----
    A linear interpolation is used when rotating the back-projection.
    This function uses ``scipy.ndimage.rotate`` for the rotation.
    """
    n_dir, l_x = projections.shape
    recons = np.zeros((l_x, l_x), dtype=float)
    angles = np.linspace(0, 180, n_dir, endpoint=False)
    for angle, line in zip(angles, projections):
        # BP: repeat the detector line along the direction of the beam
        tmp = np.tile(line[:, np.newaxis], (1, l_x))
        # Rotate the back-projection of the detector line, and add
        # it to the reconstructed image
        recons += ndimage.rotate(tmp, -angle, order=1, \
                            reshape=False)
    return recons


def projection(im, n_dir=None, interpolation='nearest'):
    """
    Tomography projection of an image along n_dir directions.

    Parameters
    ----------
    im : ndarray of square shape l_x x l_x
        Image to be projected

    n_dir : int
        Number of projection angles. Projection angles are regularly spaced
        between 0 and 180.

    interpolation : str, {'interpolation', 'nearest'}
        Interpolation method used during the projection. Default is
        'nearest'.

    Returns
    -------
    projections: ndarray of shape n_dir x l_x
        Array of projections.

    Notes
    -----
    The centers of the data pixels are projected onto the detector, then
    the contribution of a data pixel to detector pixels is computed
    by nearest neighbor or linear interpolation. The function 
    ``np.bincount`` is used to compute the projection, with weights
    corresponding to the values of data pixels, multiplied by interpolation
    weights in the case of linear interpolation.
    """
    l_x = len(im)
    if n_dir is None:
        n_dir = l_x
    im = im.ravel()
    projections = np.empty((n_dir, l_x))
    X, Y = _generate_center_coordinates(l_x)
    angles = np.linspace(0, np.pi, n_dir, endpoint=False)
    for i, angle in enumerate(angles):
        Xrot = np.cos(angle) * X - np.sin(angle) * Y 
        if interpolation == 'nearest':
            inds = _weights_nn(Xrot, dx=1, orig=X.min())
            mask = inds>= 0
            w = im[mask]
        elif interpolation == 'linear':
            inds, _, w = _weights(Xrot, dx=1, orig=X.min())
            w[:l_x**2] *= im
            w[l_x**2:] *= im
            mask = inds >= 0
            w = w[mask]
        projections[i] = np.bincount(inds[mask].astype(np.int), \
                                            weights=w)[:l_x]
    return projections

# -----------------Filtered back-projection----------------------

def filter_projections(proj_set, reg=False):
    """
    Ramp filter used in the filtered back projection.
    We use zero padding.

    Parameters
    ----------
    proj_set: 2-d ndarray
        each line is one projection (1 line of the detector) to be filtered

    Returns
    -------

    res: 2-d ndarray
        filtered projections

    Notes
    -----

    We use zero padding. However, we do not use any filtering (hanning, etc.)
    in the FFT yet.
    """
    nb_angles, l_x = proj_set.shape
    # Assum l is even for now
    ramp = 1./l_x * np.hstack((np.arange(l_x), np.arange(l_x, 0, -1)))
    return fftpack.ifft(ramp * fftpack.fft(proj_set, 2*l_x, axis=1), axis=1)[:,:l_x]




# 
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
# 
#   Copyright (C) 2011 Mirna Lerotic, 2nd Look
#   http://2ndlookconsulting.com
#   License: GNU GPL v3
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.


#    This module is used to pass data between routines and is based on 
#    https://confluence.aps.anl.gov/display/NX/Data+Exchange+Basics


class Descr(object):

    def __get__(self, instance, owner):
        # Check if the value has been set
        if (not hasattr(self, "_value")):
            return None
        # print "Getting value: %s" % self._value
        return self._value

    def __set__(self, instance, value):
        # print "Setting to %s" % value
        self._value = value

    def __delete__(self, instance):
        del (self._value)


### Information HDF5 Group
# ----------------------------------------------------------------------
class ids(object):
    proposal = Descr()
    activity = Descr()
    esaf = Descr()


# ----------------------------------------------------------------------
class experimenter(object):
    name = Descr()
    role = Descr()
    affiliation = Descr()
    address = Descr()
    phone = Descr()
    email = Descr()
    facility_user_id = Descr()


# ----------------------------------------------------------------------
class sample(object):
    name = Descr()
    description = Descr()
    #   preparation_date [string - ISO 8601 format]
    preparation_datetime = Descr()
    #   chemical_formula [string - abbreviated CIF format]
    chemical_formula = Descr()
    environment = Descr()
    temperature = Descr()
    temperature_units = Descr()
    pressure = Descr()
    pressure_units = Descr()


# ----------------------------------------------------------------------
class objective(object):
    manufacturer = Descr()
    model = Descr()
    comment = Descr()
    magnification = Descr()


# ----------------------------------------------------------------------
class scintillator(object):
    name = Descr()
    type = Descr()
    comment = Descr()
    scintillating_thickness = Descr()
    scintillating_thickness_units = Descr()
    substrate_thickness = Descr()
    substrate_thickness_units = Descr()


# ----------------------------------------------------------------------
class facility(object):
    name = Descr()
    beamline = Descr()


# ----------------------------------------------------------------------
class accelerator(object):
    ring_current = Descr()
    ring_current_units = Descr()
    primary_beam_energy = Descr()
    primary_beam_energy_units = Descr()
    monostripe = Descr()


# ----------------------------------------------------------------------
class pixel_size(object):
    horizontal = Descr()
    horizontal_units = Descr()
    vertical = Descr()
    vertical_units = Descr()


# ----------------------------------------------------------------------
class dimensions(object):
    horizontal = Descr()
    vertical = Descr()


# ----------------------------------------------------------------------
class binning(object):
    horizontal = Descr()
    vertical = Descr()


# ----------------------------------------------------------------------
class axis_directions(object):
    horizontal = Descr()
    vertical = Descr()


# ----------------------------------------------------------------------
class roi(object):
    x1 = Descr()
    y1 = Descr()
    x2 = Descr()
    y2 = Descr()


# ----------------------------------------------------------------------
class detector(object):
    manufacturer = Descr()
    model = Descr()
    serial_number = Descr()
    bit_depth = Descr()
    operating_temperature = Descr()
    operating_temperature_units = Descr()
    exposure_time = Descr()
    exposure_time_units = Descr()
    frame_rate = Descr()

    pixel_size = pixel_size()
    dimensions = dimensions()
    binning = binning()
    axis_directions = axis_directions()
    roi = roi()


# ----------------------------------------------------------------------
class information(object):
    title = Descr()
    comment = Descr()
    file_creation_datetime = Descr()

    ids = ids()
    experimenter = experimenter()
    sample = sample()
    objective = objective()
    scintillator = scintillator()
    facility = facility()
    accelerator = accelerator()
    detector = detector()


# Exchange HDF5 group
# ----------------------------------------------------------------------
class exchange(object):
    title = Descr()
    comment = Descr()
    data_collection_datetime = Descr()

    # n-dimensional dataset
    data = Descr()

    data_signal = Descr()
    data_description = Descr()
    data_units = Descr()
    data_axes = Descr()
    data_detector = Descr()

    # These are described in data attribute axes 'x:y:z' but can be arbitrary
    x = Descr()
    x_units = Descr()
    y = Descr()
    y_units = Descr()
    z = Descr()
    z_units = Descr()

    energy = Descr()
    energy_units = Descr()

    theta = Descr()
    theta_units = Descr()

    white_data = Descr()
    white_data_units = Descr()
    dark_data = Descr()
    dark_data_units = Descr()
    rotation = Descr()


# Spectromicroscopy HDF5 Group
# ----------------------------------------------------------------------
class normalization(object):
    white_spectrum = Descr()
    white_spectrum_units = Descr()
    white_spectrum_energy = Descr()
    white_spectrum_energy_units = Descr()


# ----------------------------------------------------------------------
class spectromicroscopy(object):
    positions = Descr()
    positions_units = Descr()
    positions_names = Descr()

    normalization = normalization()

    optical_density = Descr()

    data_dwell = Descr()
    i0_dwell = Descr()

    xshifts = Descr()
    yshifts = Descr()


# HDF5 Root Group
# ----------------------------------------------------------------------
class h5(object):
    # implements [string] comma separated string that tells the user which entries file contains
    implements = Descr()
    version = Descr()

    information = information()
    exchange = exchange()
    spectromicroscopy = spectromicroscopy()




# 
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
# 
#   Copyright (C) 2015 Mirna Lerotic, 2nd Look
#   http://2ndlookconsulting.com
#   License: GNU GPL v3
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.


from __future__ import division

import numpy as np
import os

          
#----------------------------------------------------------------------
def read_bim(self, filename):
    
    f = open(str(filename),'rb')
    data = np.fromfile(f, np.uint32, 6)

    nmotpos = data[0]
    ndatatype = data[1]
    ndate = data[2]
    naxisnames = data[3]

    self.n_cols = data[4]
    self.n_rows = data[5]
    self.n_ev = 1
     

    angles = np.fromfile(f, np.float64, 1)
    pixelsize = np.fromfile(f, np.float32, 1)
    
    self.x_dist = np.arange(float(self.n_cols))*pixelsize
    self.y_dist = np.arange(float(self.n_rows))*pixelsize
    
    data = np.fromfile(f, np.uint32, 2)
    
    hbin = data[0]
    vbin = data[1]
    
    energy = np.fromfile(f, np.float64, 1)
    motpos = np.fromfile(f, np.float32, nmotpos)  
    axisnames = np.fromfile(f, np.uint8, naxisnames)
    exposuretime = np.fromfile(f, np.float32, 1)
    nimages = np.fromfile(f, np.uint32, 1)
    
    data = np.fromfile(f, np.uint8, ndatatype)
    data = np.fromfile(f, np.uint8, ndate)
    
    npix = self.n_cols*self.n_rows
    
    imagestack = np.fromfile(f, np.float32, npix)
    
    self.absdata = np.reshape(imagestack, (self.n_cols, self.n_rows, self.n_ev), order='C') 
    
    
    fn = os.path.basename(str(filename))        
    fnlist = fn.split('_')
    ind = fnlist.index('eV')
    
    self.ev = [float(fnlist[ind-1])] 
      
    self.data_dwell = np.zeros((self.n_ev))+exposuretime
     

    f.close()
    

    return            
        
        
        
        
#----------------------------------------------------------------------
def read_bim_info(filename):
    
    f = open(str(filename),'rb')
    data = np.fromfile(f, np.uint32, 6)

    n_cols = data[4]
    n_rows = data[5]
   
    f.close()

    fn = os.path.basename(str(filename))        
    fnlist = fn.split('_')
    ind = fnlist.index('eV')
    
    ev = float(fnlist[ind-1])


    return n_cols, n_rows, ev     
       
        
#----------------------------------------------------------------------    
def read_bim_list(self, filelist, filepath, ds): 
    
    #Fill the common stack data 
    file1 = os.path.join(filepath, filelist[0])   
    
    f = open(str(file1),'rb') 
    data = np.fromfile(f, np.uint32, 6)

    nmotpos = data[0]
    ndatatype = data[1]
    ndate = data[2]
    naxisnames = data[3]
                
    ncols = data[4]
    nrows = data[5]
    npix = ncols*nrows
    nev = len(filelist)
    absdata = np.zeros((ncols, nrows, nev))
    ev = np.zeros((nev))
    
    
    angles = np.fromfile(f, np.float64, 1)
    pixelsize = np.fromfile(f, np.float32, 1)
    
    x_dist = np.arange(float(ncols))*pixelsize
    y_dist = np.arange(float(nrows))*pixelsize
    
    data = np.fromfile(f, np.uint32, 2)
    
    hbin = data[0]
    vbin = data[1]
    
    energy = np.fromfile(f, np.float64, 1)
    motpos = np.fromfile(f, np.float32, nmotpos)  
    axisnames = np.fromfile(f, np.uint8, naxisnames)
    exposuretime = np.fromfile(f, np.float32, 1)
    nimages = np.fromfile(f, np.uint32, 1)
    
    data = np.fromfile(f, np.uint8, ndatatype)
    data = np.fromfile(f, np.uint8, ndate)
      
    dwell_msec = exposuretime
     
    f.close()
    
   
    
    #Read the image data        
    for i in range(len(filelist)):
        fn = filelist[i]
        filename = os.path.join(filepath, fn)   

        f = open(str(filename),'rb')
        data = np.fromfile(f, np.uint32, 6)

        angles = np.fromfile(f, np.float64, 1)
        pixelsize = np.fromfile(f, np.float32, 1)
        
        data = np.fromfile(f, np.uint32, 2)
        
        energy = np.fromfile(f, np.float64, 1)
        motpos = np.fromfile(f, np.float32, nmotpos)  
        axisnames = np.fromfile(f, np.uint8, naxisnames)
        exposuretime = np.fromfile(f, np.float32, 1)
        nimages = np.fromfile(f, np.uint32, 1)
        
        data = np.fromfile(f, np.uint8, ndatatype)
        data = np.fromfile(f, np.uint8, ndate)
        
        imagestack = np.fromfile(f, np.float32, npix)
        
        absdata[:,:,i] = np.reshape(imagestack, (ncols, nrows), order='C') 
        
        fn = os.path.basename(str(filename))        
        fnlist = fn.split('_')
        ind = fnlist.index('eV')
        
        
        ev[i] = float(fnlist[ind-1])

        f.close()            
        
        
    if ev[-1]<ev[0]:
        ev = ev[::-1]
        absdata = absdata[:,:, ::-1]
        
    
    #Fill the data structure with data: 
    ds.implements = 'information:exchange:spectromicroscopy'
    ds.version = '1.0'
    ds.information.comment = 'Converted from .bim file list',
    
    import datetime 
    now = datetime.datetime.now()
    ds.information.file_creation_datetime = now.strftime("%Y-%m-%dT%H:%M")

    ds.information.experimenter.name = ''
    ds.information.sample.name = ''
    
    ds.exchange.data = absdata
    ds.exchange.data_signal = 1
    ds.exchange.data_axes='x:y'
    
    ds.exchange.energy=ev
    ds.exchange.energy_units = 'ev'
    
    
    #Since we do not have a scanning microscope we fill the x_dist and y_dist from pixel_size
    x_dist = np.arange(float(ncols))*pixelsize
    y_dist = np.arange(float(nrows))*pixelsize

    ds.exchange.x = x_dist
    ds.exchange.x_units = 'um'
    ds.exchange.y = y_dist
    ds.exchange.y_units = 'um'     
    
    
    
    self.data_dwell = np.ones((nev))*exposuretime
    ds.spectromicroscopy.data_dwell = self.data_dwell
    
    return




# 
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
# 
#   Copyright (C) 2015 Mirna Lerotic, 2nd Look
#   http://2ndlookconsulting.com
#   License: GNU GPL v3
#
#   Originally part of tomotv, code licensed under BSD license.
#   Website: https://github.com/emmanuelle/tomo-tv
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.

import numpy as np
from scipy import ndimage


def generate_synthetic_data(l_x=128, seed=None, crop=True, n_pts=25):
    """
    Generate synthetic binary data looking like phase separation

    Parameters
    ----------

    l_x: int, default 128
        Linear size of the returned image

    seed: int, default 0
        seed with which to initialize the random number generator.

    crop: bool, default True
        If True, non-zero data are found only within a central circle
        of radius l_x / 2

    n_pts: int, default 25
        number of seeds used to generate the structures. The larger n_pts,
        the finer will be the structures.

    Returns
    -------

    res: ndarray of float32, of shape lxl
        Output binary image

    Examples
    --------
    >>> im = generate_synthetic_data(l_x=256, seed=2, n_pts=25)
    >>> # Finer structures
    >>> im = generate_synthetic_data(l_x=256, n_pts=100)
    """
    if seed is None:
        seed = 0
    # Fix the seed for reproducible results
    rs = np.random.RandomState(seed)
    x, y = np.ogrid[:l_x, :l_x]
    mask = np.zeros((l_x, l_x))
    points = l_x * rs.rand(2, n_pts)
    mask[(points[0]).astype(np.int), (points[1]).astype(np.int)] = 1
    mask = ndimage.gaussian_filter(mask, sigma=l_x / (4. * np.sqrt(n_pts)))
    # Limit the non-zero data to a central circle
    if crop:
        mask_outer = (x - l_x / 2) ** 2 + (y - l_x / 2) ** 2 < (l_x / 2) ** 2
        mask = np.logical_and(mask > mask.mean(), mask_outer)
    else:
        mask = mask > mask.mean()
    return mask.astype(float32)

def tv_l0_norm(im):
    """Compute the (isotropic) TV norm of an image"""
    grad_x1 = np.diff(im, axis=0)
    grad_x2 = np.diff(im, axis=1)
    return (grad_x1[:, :-1]**2 + grad_x2[:-1, :]**2 > 0).mean()


def compute_sparsity(im):
    l_x = len(im)
    X, Y = np.ogrid[:l_x, :l_x]
    mask = ((X - l_x/2)**2 + (Y - l_x/2)**2 <= (l_x/2)**2)
    grad1 = ndimage.morphological_gradient(im, footprint=np.ones((3, 3)))
    grad2 = ndimage.morphological_gradient(im, footprint=ndimage.generate_binary_structure(2, 1))
    return (grad1[mask] > 0).mean(), (grad2[mask] > 0).mean() 



# 
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
# 
#   Copyright (C) 2015 Mirna Lerotic, 2nd Look
#   http://2ndlookconsulting.com
#   License: GNU GPL v3
#
#   Originally part of tomotv, code licensed under BSD license.
#   Website: https://github.com/emmanuelle/tomo-tv
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import numpy as np
from scipy import sparse

from .tv_denoising import tv_denoise_fista
from .projections import back_projection, projection

# ------------------ Computing energies ---------------------------

def tv_norm(im):
    """Compute the (isotropic) TV norm of an image"""
    grad_x1 = np.diff(im, axis=0)
    grad_x2 = np.diff(im, axis=1)
    return np.sqrt(grad_x1[:, :-1]**2 + grad_x2[:-1, :]**2).sum()

def tv_norm_anisotropic(im):
    """Compute the anisotropic TV norm of an image"""
    grad_x1 = np.diff(im, axis=0)
    grad_x2 = np.diff(im, axis=1)
    return np.abs(grad_x1[:, :-1]).sum() + np.abs(grad_x2[:-1, :]).sum()

# ------------------ Proximal iterators ----------------------------

def fista_tv(y, beta, niter, H, verbose=0, mask=None):
    """
    TV regression using FISTA algorithm
    (Fast Iterative Shrinkage/Thresholding Algorithm)

    Parameters
    ----------

    y : ndarray of floats
        Measures (tomography projection). If H is given, y is a column
        vector. If H is not given, y is a 2-D array where each line
        is a projection along a different angle

    beta : float
        weight of TV norm

    niter : number of forward-backward iterations to perform

    H : sparse matrix
        tomography design matrix. Should be in csr format.

    mask : array of bools

    Returns
    -------

    res : list
        list of iterates of the reconstructed images

    energies : list
        values of the function to be minimized at the different
        iterations. Its values should be decreasing.

    Notes
    -----
    This algorithm minimizes iteratively the energy

    E(x) = 1/2 || H x - y ||^2 + beta TV(x) = f(x) + beta TV(x)

    by forward - backward iterations:

    u_n = prox_{gamma beta TV}(x_n - gamma nabla f(x_n)))
    t_{n+1} = 1/2 * (1 + sqrt(1 + 4 t_n^2))
    x_{n+1} = u_n + (t_n - 1)/t_{n+1} * (u_n - u_{n-1})

    References
    ----------

    - A. Beck and M. Teboulle (2009). A fast iterative
      shrinkage-thresholding algorithm for linear inverse problems.
      SIAM J. Imaging Sci., 2(1):183-202.

    - Nelly Pustelnik's thesis (in French),
      http://tel.archives-ouvertes.fr/tel-00559126_v4/
      Paragraph 3.3.1-c p. 69 , FISTA

    """
    n_meas, n_pix = H.shape
    if mask is not None:
        l = len(mask)
    else:
        l = int(np.sqrt(n_pix))
    n_angles = n_meas / l
    Ht = sparse.csr_matrix(H.transpose())
    x0 = np.zeros(n_pix)[:, np.newaxis]
    res, energies = [], []
    gamma = .9/ (l * n_angles)
    x = x0
    u_old = np.zeros((l, l))
    t_old = 1
    for i in range(niter):
        if verbose:
            print(i)
        eps = 1.e-4
        err = H * x - y
        back_proj = Ht * err
        tmp = x - gamma * back_proj
        if mask is not None:
            tmp2d = np.zeros((l, l))
            tmp2d[mask] = tmp.ravel()
        else:
            tmp2d = tmp.reshape((l, l))
        u_n = tv_denoise_fista(tmp2d,
                weight=beta*gamma, eps=eps)
        t_new = (1 + np.sqrt(1 + 4 * t_old**2))/2.
        t_old = t_new
        x = u_n + (t_old - 1)/t_new * (u_n - u_old)
        u_old = u_n
        res.append(x)
        data_fidelity_err = 1./2 * (err**2).sum()
        tv_value = beta * tv_norm(x)
        energy = data_fidelity_err + tv_value
        energies.append(energy)
        if mask is not None:
            x = x[mask][:, np.newaxis]
        else:
            x = x.ravel()[:, np.newaxis]
    return res, energies


def ista_tv(y, beta, niter, H=None):
    """
    TV regression using ISTA algorithm
    (Iterative Shrinkage/Thresholding Algorithm)

    Parameters
    ----------

    y : ndarray of floats
        Measures (tomography projection). If H is given, y is a column
        vector. If H is not given, y is a 2-D array where each line
        is a projection along a different angle

    beta : float
        weight of TV norm

    niter : number of forward-backward iterations to perform

    H : sparse matrix or None
        tomography design matrix. Should be in csr format. If H is none,
        the projections as well as the back-projection are computed by
        a direct method, without writing explicitely the design matrix.

    Returns
    -------

    res : list
        list of iterates of the reconstructed images

    energies : list
        values of the function to be minimized at the different
        iterations. Its values should be decreasing.

    Notes
    -----

    This algorithm minimizes iteratively the energy

    E(x) = 1/2 || H x - y ||^2 + beta TV(x)

    by simple forward - backward iterations:

    x_{n + 1} = prox_{gamma beta TV(.)} (x_n - gamma nabla f(x_n))

    where f(x) = 1/2 || H x - y ||^2

    References
    ----------

    - Proximal Splitting Methods in Signal Processing, P. Combettes
      and J.-C. Pesquet, Fixed-Point Algorithms for Inverse Problems
      in Science and Engineering, p. 185 (2011). Algorithm 10.3 with
      lamba_n = 1.

    - Nelly Pustelnik's thesis (in French),
      http://tel.archives-ouvertes.fr/tel-00559126_v4/
      Paragraph 3.3.1-c p. 68 , ISTA

    """
    if H is None:
        method = 'direct'
        n_angles, l = y.shape
        n_pix = l ** 2
    else:
        method = 'matrix'
        n_angles, l = y.shape
        n_meas, n_pix = H.shape
        l = int(np.sqrt(n_pix))
        n_angles = n_meas / l
    if method == 'matrix':
        Ht = sparse.csr_matrix(H.transpose())
    x0 = np.zeros((l, l))
    res, energies = [], []
    # l * n_angles is the Lipschitz constant of Ht H
    gamma = .9/ (l * n_angles)
    x = x0
    for i in range(niter):
        eps = 1.e-4
        # Forward part
        if method == 'matrix':
            x = x.ravel()[:, np.newaxis]
            err = H * x - y
            back_proj = Ht * err
            tmp = x - gamma * back_proj
            tmp = tmp.reshape((l, l))
        else:
            err = projection(x, n_angles) - y
            back_proj = back_projection(err)
            tmp = x - gamma * back_proj
        # backward: TV prox
        x = tv_denoise_fista(tmp, weight=beta*gamma, eps=eps)
        res.append(x)
        # compute the energy
        data_fidelity_err = 1./2 * (err**2).sum()
        tv_value = beta * tv_norm(x)
        energy = data_fidelity_err + tv_value 
        energies.append(energy)
    return res, energies

def gfb_tv(y, beta, niter, H=None, val_min=0, val_max=1, x0=None,
           stop_tol=1.e-4, nonnegconst=1):
    """
    TV regression + interval constraint using the generalized
    forward backward splitting (GFB).

    Parameters
    ----------

    y : ndarray of floats
        Measures (tomography projection). If H is given, y is a column
        vector. If H is not given, y is a 2-D array where each line
        is a projection along a different angle

    beta : float
        weight of TV norm

    niter : number of forward-backward iterations to perform

    H : sparse matrix or None
        tomography design matrix. Should be in csr format. If H is none,
        the projections as well as the back-projection are computed by
        a direct method, without writing explicitely the design matrix.

    val_min, val_max: floats
        We impose that the image values are in [val_min, val_max]

    x0 : ndarray of floats, optional (default is None)
        Initial guess

    Returns
    -------

    res : list
        list of iterates of the reconstructed images

    energies : list
        values of the function to be minimized at the different
        iterations. Its values should be decreasing.

    Notes
    -----

    This algorithm minimizes iteratively the energy

    E(x) = 1/2 || H x - y ||^2 + beta TV(x) + i_C(x)

    where TV(.) is the total variation pseudo-norm and
    i_C is the indicator function of the convex set [val_min, val_max].

    The algorithm used the generalized forward-backward scheme

    z1_{n + 1} = z1_n - x_n +
                prox_{2 gamma beta TV(.)} (2*x_n - z1_n - gamma nabla f(x_n))
    z2_{n+1} =  z1_n - x_n +
                prox_{i_C(.)}(2*x_n - z2_n - gamma nabla f(x_n)

    where f(x) = 1/2 || H x - y ||^2

    This method can in fact be used for other sums of non-smooth functions
    for which the prox operator is known.

    References
    ----------
    Hugo Raguet, Jalal M. Fadili and Gabriel Peyre, Generalized
    Forward-Backward Splitting Algorithm, preprint arXiv:1108.4404v2, 2011.

    See also
    http://www.ceremade.dauphine.fr/~peyre/numerical-tour/tours/inverse_9b_gfb/
    """
    n_angles, l = y.shape
    n_meas, n_pix = H.shape
    l = int(np.sqrt(n_pix))
    n_angles = n_meas / l
    Ht = sparse.csr_matrix(H.transpose())
    if x0 is None:
        x0 = np.zeros((l, l))
    z_1 = np.zeros((l**2, 1))
    z_2 = np.zeros((l**2, 1))
    res, energies = [], []
    # l * n_angles is the Lipschitz constant of Ht H
    gamma = 2 * .9/ (l * n_angles)
    x = x0
    energy = np.inf
    for i in range(niter):
        eps = 1.e-4
        # Forward part
        x = x.ravel()[:, np.newaxis]
        err = H * x - y
        back_proj = Ht * err
        # backward: TV and i_c proxs
        # TV part
        tmp_z_1 = 2 * x - z_1 - gamma * back_proj
        tmp_z_1 = tmp_z_1.reshape((l, l))
        z_1 = z_1 + tv_denoise_fista(tmp_z_1, weight=2 * beta * gamma,
                eps=eps).ravel()[:, np.newaxis] - x
        # Projection on the interval
        tmp_z_2 = 2 * x - z_2 - gamma * back_proj
        if nonnegconst == 1: 
            tmp_z_2[tmp_z_2 < val_min] = val_min
            #print 'Non negative constraint imposed.'
        tmp_z_2[tmp_z_2 > val_max] = val_max
        z_2 = z_2 - x + tmp_z_2
        # update x: average of z_i
        x = (0.5 * (z_1 + z_2)).reshape(l, l)
        res.append(x)
        # compute the energy
        data_fidelity_err = 1./2 * (err**2).sum()
        tv_value = beta * tv_norm(x)
        energy = data_fidelity_err + tv_value
        energies.append(energy)
        # stop criterion
        if i>2 and np.abs(energy - energies[-2]) < stop_tol*energies[1]:
            break
    return res, energies


def gfb_tv_weng(y, beta, niter, H=None, val_min=0, val_max=1, x0=None,
           stop_tol=1.e-4, xb=None, xa=None, beta2=None, nonnegconst=1):
    """
    TV regression + interval constraint using the generalized
    forward backward splitting (GFB) with adjacent energies TV regularization 

    Parameters
    ----------

    y : ndarray of floats
        Measures (tomography projection). If H is given, y is a column
        vector. If H is not given, y is a 2-D array where each line
        is a projection along a different angle

    beta : float
        weight of TV norm

    niter : number of forward-backward iterations to perform

    H : sparse matrix or None
        tomography design matrix. Should be in csr format. If H is none,
        the projections as well as the back-projection are computed by
        a direct method, without writing explicitely the design matrix.

    val_min, val_max: floats
        We impose that the image values are in [val_min, val_max]

    x0 : ndarray of floats, optional (default is None)
        Initial guess at energy Em
        
    xb : ndarray of floats, optional (default is None)
        Initial guess at adjacent energy Em-1
        
    xa : ndarray of floats, optional (default is None)
        Initial guess at adjacent energy Em+1

    Returns
    -------

    res : list
        list of iterates of the reconstructed images

    CSenergies : list
        values of the function to be minimized at the different
        iterations. Its values should be decreasing.

    Notes
    -----

    This algorithm minimizes iteratively the CSenergy

    E(x) = 1/2 || H x - y ||^2 + beta TV(x) + +beta2 TV(x,xa,xb) + i_C(x)

    where TV(.) is the total variation pseudo-norm and
    i_C is the indicator function of the convex set [val_min, val_max].

    The algorithm used the generalized forward-backward scheme

    z1_{n + 1} = z1_n - x_n +
                prox_{2 gamma beta TV(.)} (2*x_n - z1_n - gamma nabla f(x_n))
    z2_{n+1} =  z1_n - x_n +
                prox_{i_C(.)}(2*x_n - z2_n - gamma nabla f(x_n))
                
    z3_{n + 1} = z1_n - x_n + 0.1 (xa+xb) + 
                prox_{2 gamma beta2 TV(.)} (2*x_n - z1_n - gamma nabla f(x_n))
                


    where f(x) = 1/2 || H x - y ||^2

    This method can in fact be used for other sums of non-smooth functions
    for which the prox operator is known.

    References
    ----------
    Hugo Raguet, Jalal M. Fadili and Gabriel Peyre, Generalized
    Forward-Backward Splitting Algorithm, preprint arXiv:1108.4404v2, 2011.

    See also
    http://www.ceremade.dauphine.fr/~peyre/numerical-tour/tours/inverse_9b_gfb/
    """
    n_angles, l = y.shape
    n_meas, n_pix = H.shape
    l = int(np.sqrt(n_pix))
    n_angles = n_meas / l
    Ht = sparse.csr_matrix(H.transpose())
    if x0 is None:
        x0 = np.zeros((l, l))
    z_1 = np.zeros((l**2, 1))
    z_2 = np.zeros((l**2, 1))
    z_3 = np.zeros((l**2, 1))
    res, energies = [], []
    xb = xb.ravel()[:, np.newaxis]
    xa = xa.ravel()[:, np.newaxis]
    if beta2 == None:
        #beta2=beta
        beta2=0
    # l * n_angles is the Lipschitz constant of Ht H
    gamma = 2 * .9/ (l * n_angles)
    x = x0
    energy = np.inf
    for i in range(niter):
        eps = 1.e-4
        # Forward part
        x = x.ravel()[:, np.newaxis]
        err = H * x - y
        back_proj = Ht * err
        # backward: TV and i_c proxs
        # TV part
        tmp_z_1 = 2 * x - z_1 - gamma * back_proj
        tmp_z_1 = tmp_z_1.reshape((l, l))
        z_1 = z_1 + tv_denoise_fista(tmp_z_1, weight=2 * beta * gamma,
                eps=eps).ravel()[:, np.newaxis] - x
        # Projection on the interval
        tmp_z_2 = 2 * x - z_2 - gamma * back_proj
        if nonnegconst:
            tmp_z_2[tmp_z_2 < val_min] = val_min
        tmp_z_2[tmp_z_2 > val_max] = val_max
        z_2 = z_2 - x + tmp_z_2
        if beta2 != 0:
            #Use images taken at energy before and after the current one
            tmp_z_3 = 2 * x - z_3 - gamma * back_proj
            tmp_z_3 = (tmp_z_3+(xa+xb)*.1).reshape((l, l))
            z_3 = z_3 + tv_denoise_fista(tmp_z_3, weight=2 * beta2 * gamma,
                    eps=eps).ravel()[:, np.newaxis] - x       
            # update x: average of z_i
            x = (0.3333 * (z_1 + z_2 + z_3)).reshape(l, l)
        else:
            x = (0.5 * (z_1 + z_2)).reshape(l, l)
        res.append(x)
        # compute the energy
        data_fidelity_err = 1./2 * (err**2).sum()
        tv_value = beta * tv_norm(x)
        energy = data_fidelity_err + tv_value
        energies.append(energy)
        # stop criterion
        if i>2 and np.abs(energy - energies[-2]) < stop_tol*energies[1]:
            break
    
    return res, energies

def gfb_tv_local(y, beta, niter, mask_pix, mask_reg, H=None,
                                val_min=0, val_max=1, x0=None):
    """
    TV regression + interval constraint using the generalized
    forward backward splitting (GFB), in local tomography mode.

    Parameters
    ----------

    y : ndarray of floats
        Measures (tomography projection). If H is given, y is a column
        vector. If H is not given, y is a 2-D array where each line
        is a projection along a different angle

    beta : float
        weight of TV norm

    niter : number of forward-backward iterations to perform

    mask_pix: ndarray of bools
        Domain where pixels are reconstructed (typically, the disk
        inside a square).

    mask_reg: ndarray of bools
        Domain where the spatial regularization is performed

    H : sparse matrix or None
        tomography design matrix. Should be in csr format. If H is none,
        the projections as well as the back-projection are computed by
        a direct method, without writing explicitely the design matrix.

    val_min, val_max: floats
        We impose that the image values are in [val_min, val_max]

    x0 : ndarray of floats, optional (default is None)
        Initial guess

    Returns
    -------

    res : list
        list of iterates of the reconstructed images

    energies : list
        values of the function to be minimized at the different
        iterations. Its values should be decreasing.

    Notes
    -----

    This algorithm minimizes iteratively the energy

    E(x) = 1/2 || H x - y ||^2 + beta TV(x) + i_C(x)

    where TV(.) is the total variation pseudo-norm and
    i_C is the indicator function of the convex set [val_min, val_max].

    The algorithm used the generalized forward-backward scheme

    z1_{n + 1} = z1_n - x_n +
                prox_{2 gamma beta TV(.)} (2*x_n - z1_n - gamma nabla f(x_n))
    z2_{n+1} =  z1_n - x_n +
                prox_{i_C(.)}(2*x_n - z2_n - gamma nabla f(x_n)

    where f(x) = 1/2 || H x - y ||^2

    This method can in fact be used for other sums of non-smooth functions
    for which the prox operator is known.

    References
    ----------
    Hugo Raguet, Jalal M. Fadili and Gabriel Peyre, Generalized
    Forward-Backward Splitting Algorithm, preprint arXiv:1108.4404v2, 2011.

    See also
    http://www.ceremade.dauphine.fr/~peyre/numerical-tour/tours/inverse_9b_gfb/
    """
    mask_reg = mask_reg[mask_pix]
    n_meas, n_pix = H.shape
    l = len(mask_pix)
    n_angles = n_meas / l
    Ht = sparse.csr_matrix(H.transpose())
    z_1 = np.zeros((n_pix, 1))
    z_2 = np.zeros((n_pix, 1))
    res, energies = [], []
    # l * n_angles is the Lipschitz constant of Ht H
    gamma = 2 * .5/ (l * n_angles)
    x0 = np.zeros(n_pix)[:, np.newaxis]
    x = x0
    for i in range(niter):
        eps = 1.e-4
        # Forward part
        err = H * x - y
        back_proj = Ht * err
        grad_descent = x - gamma * back_proj
        # backward: TV and i_c proxs
        # TV part
        tmp_z_1 = 2 * x - z_1 - gamma * back_proj
        tmp_z_1_2d = np.zeros((l, l))
        tmp_z_1_2d[mask_pix] = tmp_z_1.ravel()
        z_1 = z_1 + tv_denoise_fista(tmp_z_1_2d, weight=2 * beta * gamma,
                eps=eps)[mask_pix][:, np.newaxis] - x
        # Projection on the interval
        tmp_z_2 = 2 * x - z_2 - gamma * back_proj
        tmp_z_2[tmp_z_2 < val_min] = val_min
        tmp_z_2[tmp_z_2 > val_max] = val_max
        z_2 = z_2 - x + tmp_z_2
        # update x: average of z_i
        x = (0.5 * (z_1 + z_2))
        x[~mask_reg] = grad_descent[~mask_reg]
        tmp = np.zeros((l, l))
        tmp[mask_pix] = x.ravel()
        res.append(tmp)
    return res, energies



#
#   This file is part of Mantis, a Multivariate ANalysis Tool for Spectromicroscopy.
#
#   Copyright (C) 2013 Mirna Lerotic, 2nd Look
#   http://2ndlookconsulting.com
#   License: GNU GPL v3
#
#   Mantis is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   any later version.
#
#   Mantis is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details <http://www.gnu.org/licenses/>.


import sys
import os
import numpy as np
import matplotlib
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
#from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import colorbar

#Internal imports
from . import data_struct
from . import data_stack
from . import analyze
from . import nnma
from . import henke
from . import tomo_reconstruction

from . import file_plugins
from .file_plugins import file_xrm
from .file_plugins import file_bim
from .file_plugins import file_dataexch_hdf5
from .file_plugins import file_ncb
from .file_plugins import file_json
from .file_plugins import file_tif
from .file_plugins import file_stk
from .file_plugins import file_csv
    

PlotH = 4.0
PlotW = PlotH*1.61803



#----------------------------------------------------------------------
def save_keyeng(key_engs, odir, filename, stk, anlz, png, pdf, svg):

    SaveFileName = os.path.join(odir,filename)

    matplotlib.rcParams['pdf.fonttype'] = 42

    fig = matplotlib.figure.Figure(figsize =(PlotW, PlotH))
    canvas = FigureCanvas(fig)
    fig.clf()
    fig.add_axes((0.15,0.15,0.75,0.75))
    axes = fig.gca()

    odtotal = stk.od3d.sum(axis=0)
    odtotal = odtotal.sum(axis=0)/(stk.n_rows*stk.n_cols)
    odtotal /= odtotal.max()/0.7

    specplot = axes.plot(stk.ev,odtotal)

    for i in range(len(key_engs)):
        axes.axvline(x=key_engs[i], color = 'g', alpha=0.5)

    axes.set_xlabel('Photon Energy [eV]')
    axes.set_ylabel('Optical Density')

    if png == 1:
        ext = 'png'
        fileName_img = SaveFileName+"_keyengs."+ext
        fig.savefig(fileName_img, pad_inches = 0.0)
    if pdf == 1:
        ext = 'pdf'
        fileName_img = SaveFileName+"_keyengs."+ext
        fig.savefig(fileName_img, pad_inches = 0.0)
    if svg == 1:
        ext = 'svg'
        fileName_img = SaveFileName+"_keyengs."+ext
        fig.savefig(fileName_img, pad_inches = 0.0)


    #Save text file with list of energies
    textfilepath = SaveFileName+'_keyenergies.csv'
    f = open(textfilepath, 'w')
    print('*********************  Key Energies  ********************', file=f)
    for i in range(len(key_engs)):
        print('%.6f' %(key_engs[i]), file=f)

    f.close()

    return


#----------------------------------------------------------------------
def save_spa(odir, filename, stk, anlz, png, pdf, svg):

    SaveFileName = os.path.join(odir,filename)

    matplotlib.rcParams['pdf.fonttype'] = 42

    colors=['#FF0000','#000000','#FFFFFF']
    spanclrmap=matplotlib.colors.LinearSegmentedColormap.from_list('spancm',colors)

    for i in range (anlz.n_target_spectra):

        #Save composition maps
        if anlz.pca_calculated == 0:
            tsmapimage = anlz.target_svd_maps[:,:,i]
        else:
            tsmapimage = anlz.target_pcafit_maps[:,:,i]

        fig = matplotlib.figure.Figure(figsize =(PlotH, PlotH))
        canvas = FigureCanvas(fig)
        fig.clf()
        axes = fig.gca()

        divider = make_axes_locatable(axes)
        ax_cb = divider.new_horizontal(size="3%", pad=0.03)

        fig.add_axes(ax_cb)
        axes.set_position([0.03,0.03,0.8,0.94])

        min_val = np.min(tsmapimage)
        max_val = np.max(tsmapimage)
        bound = np.max((np.abs(min_val), np.abs(max_val)))


        im = axes.imshow(tsmapimage, cmap=spanclrmap, vmin = -bound, vmax = bound)
        cbar = axes.figure.colorbar(im, orientation='vertical',cax=ax_cb)

        axes.axis("off")

        if png == 1:
            ext = 'png'
            fileName_img = SaveFileName+"_TSmap_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, pad_inches = 0.0)
        if pdf == 1:
            ext = 'pdf'
            fileName_img = SaveFileName+"_TSmap_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, pad_inches = 0.0)
        if svg == 1:
            ext = 'svg'
            fileName_img = SaveFileName+"_TSmap_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, pad_inches = 0.0)

    #Save spectra
    for i in range (anlz.n_target_spectra):

        tspectrum = anlz.target_spectra[i, :]

        fig = matplotlib.figure.Figure(figsize =(PlotW, PlotH))
        canvas = FigureCanvas(fig)
        fig.clf()
        fig.add_axes((0.15,0.15,0.75,0.75))
        axes = fig.gca()

        line1 = axes.plot(stk.ev,tspectrum, color='black', label = 'Raw data')

        if anlz.pca_calculated == 1:
            tspectrumfit = anlz.target_pcafit_spectra[i, :]
            diff = np.abs(tspectrum-tspectrumfit)
            line2 = axes.plot(stk.ev,tspectrumfit, color='green', label = 'Fit')
            line3 = axes.plot(stk.ev,diff, color='grey', label = 'Abs(Raw-Fit)')

        fontP = matplotlib.font_manager.FontProperties()
        fontP.set_size('small')

        axes.legend(loc=4, prop = fontP)

        axes.set_xlabel('Photon Energy [eV]')
        axes.set_ylabel('Optical Density')

        if png == 1:
            ext = 'png'
            fileName_spec = SaveFileName+"_Tspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)
        if pdf == 1:
            ext = 'pdf'
            fileName_spec = SaveFileName+"_Tspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)
        if svg == 1:
            ext = 'svg'
            fileName_spec = SaveFileName+"_Tspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)

        fileName_spec = SaveFileName+"_Tspectrum_" +str(i+1)+".csv"
        cname = "Tspectrum_" +str(i+1)
        stk.write_csv(fileName_spec, stk.ev, tspectrum, cname = cname)

    return


#----------------------------------------------------------------------
def MakeColorTable():
    maxclcolors = 11
    colors_i = np.linspace(0, maxclcolors, maxclcolors+1)

    colors=['#0000FF','#FF0000','#DFE32D','#36F200','#B366FF',
            '#FF470A','#33FFFF','#006600','#CCCC99','#993300',
            '#000000']

    clusterclrmap1=matplotlib.colors.LinearSegmentedColormap.from_list('clustercm',colors)

    bnorm1 = matplotlib.colors.BoundaryNorm(colors_i, clusterclrmap1.N)

    colors_i = np.linspace(0,maxclcolors+2,maxclcolors+3)

    #use black color for clusters > maxclcolors, the other 2 colors are for background
    colors2=['#0000FF','#FF0000','#DFE32D','#36F200','#B366FF',
            '#FF470A','#33FFFF','#006600','#CCCC99','#993300',
            '#000000','#FFFFFF','#EEEEEE']

    clusterclrmap2=matplotlib.colors.LinearSegmentedColormap.from_list('clustercm2',colors2)

    bnorm2 = matplotlib.colors.BoundaryNorm(colors_i, clusterclrmap2.N)

    return clusterclrmap1, bnorm1, clusterclrmap2, bnorm2


#----------------------------------------------------------------------
#If png_pdg = 1 save png, if =2 save pdf, if =3 save svg
def SaveScatt(SaveFileName, stk, anlz, png_pdf = 1):

    colors=['#0000FF','#FF0000','#DFE32D','#36F200','#B366FF',
            '#FF470A','#33FFFF','#006600','#CCCC99','#993300',
            '#000000']

    od_reduced = anlz.pcaimages[:,:,0:anlz.numsigpca]
    od_reduced = np.reshape(od_reduced, (stk.n_cols*stk.n_rows,anlz.numsigpca), order='F')

    clindices = anlz.cluster_indices
    clindices = np.reshape(clindices, (stk.n_cols*stk.n_rows), order='F')

    if png_pdf == 1:
        ext = 'png'
    elif png_pdf == 2:
        ext = 'pdf'
    elif png_pdf == 3:
        ext = 'svg'

    nplots = 0
    for ip in range(anlz.numsigpca):
        for jp in range(anlz.numsigpca):
            if jp >= (ip+1):
                nplots = nplots+1
    nplotsrows = np.ceil(nplots/2)

    plotsize = 2.5

    if nplots > 1 :
        fig = matplotlib.figure.Figure(figsize =(6.0,plotsize*nplotsrows))
        fig.subplots_adjust(wspace = 0.4, hspace = 0.4)
    else:
        fig = matplotlib.figure.Figure(figsize =(3.0,2.5))
        fig.subplots_adjust(bottom = 0.2, left = 0.2)

    canvas = FigureCanvas(fig)
    #axes = fig.gca()
    matplotlib.rcParams['font.size'] = 6

    pplot = 1
    for ip in range(anlz.numsigpca):
        for jp in range(anlz.numsigpca):
            if jp >= (ip+1):


                x_comp = od_reduced[:,ip]
                y_comp = od_reduced[:,jp]
                if nplots > 1 :
                    axes = fig.add_subplot(nplotsrows,2, pplot)
                else:
                    axes = fig.add_subplot(1,1,1)

                pplot = pplot+1

                for i in range(anlz.nclusters):
                    thiscluster = np.where(clindices == i)
                    axes.plot(x_comp[thiscluster], y_comp[thiscluster],'.',color=colors[i],alpha=0.5)
                axes.set_xlabel('Component '+str(ip+1))
                axes.set_ylabel('Component '+str(jp+1))

    fileName_sct = SaveFileName+"_CAscatterplots."+ext
    matplotlib.rcParams['pdf.fonttype'] = 42
    fig.savefig(fileName_sct)


#----------------------------------------------------------------------
def save_ca(odir, filename, stk, anlz, png, pdf, svg):

    clusterclrmap1, bnorm1, clusterclrmap2, bnorm2 = MakeColorTable()
    maxclcolors = 11
    colors=['#0000FF','#FF0000','#DFE32D','#36F200','#B366FF',
            '#FF470A','#33FFFF','#006600','#CCCC99','#993300',
            '#000000']

    SaveFileName = os.path.join(odir,filename)

    matplotlib.rcParams['pdf.fonttype'] = 42

    #Save image with composite cluster indices
    fig = matplotlib.figure.Figure(figsize = (float(stk.n_rows)/30, float(stk.n_cols)/30))
    canvas = FigureCanvas(fig)
    fig.clf()
    fig.add_axes((0.0,0.0,1.0,1.0))
    axes = fig.gca()

    im = axes.imshow(anlz.cluster_indices, cmap=clusterclrmap1, norm=bnorm1)
    axes.axis("off")

    if png == 1:
        ext = 'png'
        fileName_caimg = SaveFileName+"_CAcimg."+ext
        fig.savefig(fileName_caimg, dpi=300, pad_inches = 0.0)
    if pdf == 1:
        ext = 'pdf'
        fileName_caimg = SaveFileName+"_CAcimg."+ext
        fig.savefig(fileName_caimg, dpi=300, pad_inches = 0.0)
    if svg == 1:
        ext = 'svg'
        fileName_caimg = SaveFileName+"_CAcimg."+ext
        fig.savefig(fileName_caimg, dpi=300, pad_inches = 0.0)

    #Save individual cluster images
    for i in range (anlz.nclusters):

        indvclusterimage = np.zeros((anlz.stack.n_cols, anlz.stack.n_rows))+20.
        ind = np.where(anlz.cluster_indices == i)
        colorcl = min(i,9)
        indvclusterimage[ind] = colorcl

        fig = matplotlib.figure.Figure(figsize =(float(stk.n_rows)/30, float(stk.n_cols)/30))
        canvas = FigureCanvas(fig)
        fig.add_axes((0.0,0.0,1.0,1.0))
        axes = fig.gca()
        im = axes.imshow(indvclusterimage, cmap=clusterclrmap2, norm=bnorm2)
        axes.axis("off")

        if png == 1:
            ext = 'png'
            fileName_img = SaveFileName+"_CAimg_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, dpi=300, pad_inches = 0.0)
        if pdf == 1:
            ext = 'pdf'
            fileName_img = SaveFileName+"_CAimg_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, dpi=300, pad_inches = 0.0)
        if svg == 1:
            ext = 'svg'
            fileName_img = SaveFileName+"_CAimg_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, dpi=300, pad_inches = 0.0)

    for i in range (anlz.nclusters):

        clusterspectrum = anlz.clusterspectra[i, ]
        fig = matplotlib.figure.Figure(figsize =(PlotW, PlotH))
        canvas = FigureCanvas(fig)
        fig.add_axes((0.15,0.15,0.75,0.75))
        axes = fig.gca()
        if i >= maxclcolors:
            clcolor = colors[maxclcolors-1]
        else:
            clcolor = colors[i]

        specplot = axes.plot(anlz.stack.ev,clusterspectrum, color = clcolor)

        axes.set_xlabel('Photon Energy [eV]')
        axes.set_ylabel('Optical Density')

        if png == 1:
            ext = 'png'
            fileName_spec = SaveFileName+"_CAspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)
        if pdf == 1:
            ext = 'pdf'
            fileName_spec = SaveFileName+"_CAspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)
        if svg == 1:
            ext = 'svg'
            fileName_spec = SaveFileName+"_CAspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)

        #Save all spectra in one plot
        fig = matplotlib.figure.Figure(figsize =(PlotW, PlotH))
        canvas = FigureCanvas(fig)
        fig.add_axes((0.15,0.15,0.75,0.75))
        axes = fig.gca()

        for i in range (anlz.nclusters):

            clusterspectrum = anlz.clusterspectra[i-1, ]/np.amax(anlz.clusterspectra[i-1, ])

            if i >= maxclcolors:
                clcolor = colors[maxclcolors-1]
            else:
                clcolor = colors[i]

            specplot = axes.plot(anlz.stack.ev,clusterspectrum, color = clcolor)

            axes.set_xlabel('Photon Energy [eV]')
            axes.set_ylabel('Optical Density')
        if png == 1:
            ext = 'png'
            fileName_spec = SaveFileName+"_CAspectra"+"."+ext
            fig.savefig(fileName_spec)
        if pdf == 1:
            ext = 'pdf'
            fileName_spec = SaveFileName+"_CAspectra"+"."+ext
            fig.savefig(fileName_spec)
        if svg == 1:
            ext = 'svg'
            fileName_spec = SaveFileName+"_CAspectra"+"."+ext
            fig.savefig(fileName_spec)

    for i in range (anlz.nclusters):
        clusterspectrum = anlz.clusterspectra[i, ]
        fileName_spec = SaveFileName+"_CAspectrum_" +str(i+1)+".csv"
        cname = "CAspectrum_" +str(i+1)
        stk.write_csv(fileName_spec, anlz.stack.ev, clusterspectrum, cname=cname)

    if png:
        SaveScatt(SaveFileName, stk, anlz, png_pdf = 1)
    if pdf:
        SaveScatt(SaveFileName, stk, anlz, png_pdf = 2)
    if svg:
        SaveScatt(SaveFileName, stk, anlz, png_pdf = 3)

    return


#----------------------------------------------------------------------
def save_pca(odir, filename, stk, anlz, png, pdf, svg):

    SaveFileName = os.path.join(odir,filename)

    matplotlib.rcParams['pdf.fonttype'] = 42

    #Save evals
    evalmax = np.min([stk.n_ev, 40])
    pcaevals = anlz.eigenvals[0:evalmax]
    fig = Figure(figsize =(PlotW, PlotH))
    canvas = FigureCanvas(fig)

    fig.clf()
    fig.add_axes((0.15,0.15,0.75,0.75))
    axes = fig.gca()

    evalsplot = axes.semilogy(np.arange(1,evalmax+1), pcaevals,'b.')

    axes.set_xlabel('Principal Component')
    axes.set_ylabel('Log(Eigenvalue)')

    if png == 1:
        ext = 'png'
        fileName_evals = SaveFileName+"_PCAevals."+ext
        fig.savefig(fileName_evals, pad_inches = 0.0)
    if pdf == 1:
        ext = 'pdf'
        fileName_evals = SaveFileName+"_PCAevals."+ext
        fig.savefig(fileName_evals, pad_inches = 0.0)
    if svg == 1:
        ext = 'svg'
        fileName_evals = SaveFileName+"_PCAevals."+ext
        fig.savefig(fileName_evals, pad_inches = 0.0)

    for i in range(10):
        pcaimage = anlz.pcaimages[:,:,i]

        fig = Figure(figsize =(PlotH*1.15, PlotH))
        canvas = FigureCanvas(fig)
        axes = fig.gca()
        divider = make_axes_locatable(axes)
        ax_cb = divider.new_horizontal(size="3%", pad=0.03)
        fig.add_axes(ax_cb)
        axes.set_position([0.03,0.03,0.8,0.94])
        bound = anlz.pcaimagebounds[i]

        im = axes.imshow(pcaimage, cmap=matplotlib.cm.get_cmap("seismic_r"), vmin = -bound, vmax = bound)

        cbar = colorbar(im, orientation='vertical', cax=ax_cb)
        axes.axis("off")

        if png == 1:
            ext = 'png'
            fileName_img = SaveFileName+"_PCA_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, pad_inches = 0.0)
        if pdf == 1:
            ext = 'pdf'
            fileName_img = SaveFileName+"_PCA_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, pad_inches = 0.0)
        if svg == 1:
            ext = 'svg'
            fileName_img = SaveFileName+"_PCA_" +str(i+1)+"."+ext
            fig.savefig(fileName_img, pad_inches = 0.0)

    for i in range(10):
        pcaspectrum = anlz.eigenvecs[:,i]
        fig = Figure(figsize =(PlotW, PlotH))
        canvas = FigureCanvas(fig)
        fig.add_axes((0.15,0.15,0.75,0.75))
        axes = fig.gca()
        specplot = axes.plot(stk.ev, pcaspectrum)
        axes.set_xlabel('Photon Energy [eV]')
        axes.set_ylabel('Optical Density')
        if png == 1:
            ext = 'png'
            fileName_spec = SaveFileName+"_PCAspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)
        if pdf == 1:
            ext = 'pdf'
            fileName_spec = SaveFileName+"_PCAspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)
        if svg == 1:
            ext = 'svg'
            fileName_spec = SaveFileName+"_PCAspectrum_" +str(i+1)+"."+ext
            fig.savefig(fileName_spec)

    for i in range (10):
        pcaspectrum = anlz.eigenvecs[:,i]
        fileName_spec = SaveFileName+"_PCAspectrum_" +str(i+1)+".csv"
        cname = "PCAspectrum_" +str(i+1)
        stk.write_csv(fileName_spec, stk.ev, pcaspectrum, cname=cname)

    return


#----------------------------------------------------------------------
def batch_mode():

    verbose  = 1

    settingsfile = 'Mantis_batch_settings.txt'

    version = '2.0.5'
    wdir = ''
    outdir = 'MantisResults'
    filename = ''
    save_hdf5 = 0
    align_stack = 0
    i0_file = ''
    i0_histogram = 0
    run_pca = 0
    n_spca = 4
    run_ca = 0
    nclusters = 5
    ca_thickness = 0
    run_sa = 0
    sa_spectra = []
    sa_use_clspectra = 0
    run_keyengs = 0
    kengs_thresh = 0.10
    save_png = 1
    save_pdf = 0
    save_svg = 0

    try:
        f = open(settingsfile, 'rt')
        for line in f:
            if ':' in line :
                slist = line.split(':')
                tag = slist[0]
                value = ':'.join(slist[1:])

                if   tag == 'VERSION': version = float(value)
                elif tag == 'WORK_DIR' : wdir  =  value.strip()
                elif tag == 'OUTPUT_DIR_NAME' : outdir  =  value.strip()
                elif tag == 'FILENAME' : filename  =  value.strip()
                elif tag == 'ALIGN_STACK' : align_stack  =  value.strip()
                elif tag == 'I0_FILE' : i0_file  =  value.strip()
                elif tag == 'I0_HISTOGRAM' : i0_histogram = int(value)
                elif tag == 'SAVE_HDF5' : save_hdf5 = int(value)
                elif tag == 'RUN_PCA' : run_pca = int(value)
                elif tag == 'N_SPCA' : n_spca = int(value)
                elif tag == 'RUN_CLUSTER_ANALYSIS' : run_ca = int(value)
                elif tag == 'N_CLUSTERS' : nclusters = int(value)
                elif tag == 'THICKNESS_CORRECTION' : ca_thickness = int(value)
                elif tag == 'RUN_SPECTRAL_ANALYSIS' : run_sa = int(value)
                elif tag == 'SA_SPECTRUM' :
                    spname = value.strip()
                    if len(spname) > 0 :
                        sa_spectra.append(spname)
                elif tag == 'SA_USE_CA_SPECTRA' : sa_use_clspectra = int(value)
                elif tag == 'RUN_KEY_ENGS' : run_keyengs = int(value)
                elif tag == 'KE_THRESHOLD' : kengs_thresh = float(value)
                elif tag == 'SAVE_PNG' : save_png = int(value)
                elif tag == 'SAVE_PDF' : save_pdf = int(value)
                elif tag == 'SAVE_SVG' : save_svg = int(value)

        f.close()

    except:
        print('Error: Could not read in Mantis_batch_settings.txt.')
        return

    wdir = os.path.normpath(wdir)

    if verbose:
        print('Version: ', version)
        print('Working directory: ', wdir)

    if not os.path.exists(wdir):
        print('Error - Directory ', wdir, ' does not exist. Please specify working directory.')
        return

    outdir = os.path.join(wdir, outdir)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
        if not os.path.exists(outdir):
            print('Error: Did not find and could not create a new output directory.')
            return

    if save_png == 1:
        print("Save .png images")

    if save_pdf == 1:
        print("Save .pdf images")

    datastruct = data_struct.h5()
    stk = data_stack.data(datastruct)
    anlz = analyze.analyze(stk)

    print('Reading file:', filename)
    basename, extension = os.path.splitext(filename)
    filepath = os.path.join(wdir, filename)

    supported_filters = file_plugins.supported_filters
    filter_list = file_plugins.filter_list

    action = 'read'
    data_type = 'stack'

    print (filter_list[action][data_type])

    plugin = file_plugins.identify(filepath)

    file_plugins.load(filepath, stack_object=stk, plugin=plugin, selection=None, json=None)


    if align_stack:
        print('Aligning the stack')
        xshifts = np.zeros((stk.n_ev))
        yshifts = np.zeros((stk.n_ev))

        referenceimage = stk.absdata[:,:,0].copy()

        for i in range(stk.n_ev):

            img2 = stk.absdata[:,:,i]

            if i==0:
                xshift, yshift, ccorr = stk.register_images(referenceimage, img2,
                                                          have_ref_img_fft = False)
            else:
                xshift, yshift, ccorr = stk.register_images(referenceimage, img2,
                                                          have_ref_img_fft = True)

#             #Limit the shifts to MAXSHIFT chosen by the user
#             if (self.maxshift > 0):
#                 if (abs(xshift) > self.maxshift):
#                         xshift = npy.sign(xshift)*self.maxshift
#                 if (abs(yshift) > self.maxshift):
#                         yshift = npy.sign(yshift)*self.maxshift

            xshifts[i] = xshift
            yshifts[i] = yshift

        #Apply shifts
        for i in range(stk.n_ev):
            img = stk.absdata[:,:,i]
            if (abs(xshifts[i])>0.02) or (abs(yshifts[i])>0.02):
                shifted_img = stk.apply_image_registration(img, xshifts[i], yshifts[i])
                stk.absdata[:,:,i] = shifted_img

    if datastruct.spectromicroscopy.normalization.white_spectrum is not None:
        print("I0 loaded")
    else:
        print("Loading I0")
        if i0_histogram == 1:
            print('Getting I0 from the histogram')
            stk.calc_histogram()
            averagefluxmax = np.max(stk.histogram)
            histmin = 0.98*averagefluxmax
            histmax = averagefluxmax
            stk.i0_from_histogram(histmin, histmax)

        elif len(i0_file) > 0:
            print('Reading I0 from file:', i0_file)
            i0basename, i0extension = os.path.splitext(i0_file)
            i0filepath = os.path.join(wdir, i0_file)
            stk.read_stk_i0(i0filepath, i0extension)

        else:
            print("Please either set I0_HISTOGRAM to 1 to calculate I0 or specify I0 file.")
            return

    if datastruct.spectromicroscopy.normalization.white_spectrum is None:
        print('Error: I0 not loaded')
        return

    if save_hdf5 == 1:
        fnameh5 =  os.path.join(wdir,basename+'_MantisBatch.hdf5')
        stk.write_h5(fnameh5, data_struct)
        print('Saving data to HDF5 file:', fnameh5)

    pca_calculated = 0
    if run_pca == 1:
        print("Running PCA Analysis")
        anlz.calculate_pca()
        print("Chosen number of significant components:", n_spca)
        print("Suggested number of significant components:", anlz.numsigpca)
        pca_calculated = 1
        anlz.numsigpca = n_spca
        save_pca(outdir, filename, stk, anlz, save_png, save_pdf, save_svg)

    ca_calculated = 0
    if run_ca == 1:
        if pca_calculated == 0:
            anlz.calculate_pca()
        print("Running Cluster Analysis")
        print("Number of clusters",  nclusters)
        if ca_thickness == 1:
            print("Thickness correction enabled")
        nclusters = anlz.calculate_clusters(nclusters, ca_thickness)
        ca_calculated = 1
        save_ca(outdir, filename, stk, anlz, save_png, save_pdf, save_svg)

    if run_sa == 1:
        print("Running Spectral Analysis")
        if len(sa_spectra) > 0:
            print("Loading spectra:", sa_spectra)
            for i in range(len(sa_spectra)):
                sppath = os.path.join(wdir, sa_spectra[i])
                anlz.read_target_spectrum(filename=sppath)

        if sa_use_clspectra == 1:
            if ca_calculated == 1:
                print("Loading cluster spectra")
                anlz.add_cluster_target_spectra()
            else:
                print("Please set RUN_CLUSTER_ANALYSIS to 1 to calculate cluster spectra.")

        if anlz.n_target_spectra > 1:
            save_spa(outdir, filename, stk, anlz, save_png, save_pdf, save_svg)

    if run_keyengs == 1:
        if pca_calculated == 0:
            anlz.calculate_pca()
        print("Finding key energies")
        print("Threshold for finding key energies:", kengs_thresh)
        key_engs= anlz.calc_key_engs(kengs_thresh)
        save_keyeng(key_engs, outdir, filename, stk, anlz, save_png, save_pdf, save_svg)

    if (save_hdf5 == 1) and (pca_calculated == 1) :
        fnameh5 =  os.path.join(wdir,basename+'_MantisBatch.hdf5')
        stk.write_results_h5(fnameh5, data_struct, anlz)


    print("Finished doing Mantis analysis")
    return


""" ------------------------------------------------------------------------------------------------"""
def main():

    verbose = True

    print('Running Mantis in batch mode.')
    batch_mode()

    sys.exit()

if __name__ == '__main__':
    main()

